; ### $B ###
; Port of the game 2048 for unexpanded VIC-20.

; # const

AREA_W = 4
AREA_H = AREA_W
AREA_SIZE = AREA_W * AREA_H

TOK_CLEAR = $93
TOK_HOME = $13
TOK_CR = $0d

LAST_KEY = 4


; # addresses

score = $00
scoreh = $01
; move command; 0..3 -> LRUD
;   bit 0: mirror
;   bit 1: rotate
cmd = $02
tmp = $03
newrow = $04    ; .. $07
newrowend = $08 ; zero termination
areachanges = $09   ; >0 -> area changed during move
newrow2 = $69

; 4x4 play area
area = $35  ; .. $44


b_linprt = $ddcd  ; print a:x in decimal
b_intidx = $d1aa   ; fac to a:y
b_y_to_fac = $d3a2
b_mul_ay_fac = $da30
b_rftoa = $dc0c
k_rnd_1_to_fac = $e0bb
k_getinkbd = $f1f9 ;$ffe4
k_chrout = $f27a ;$ffd2


; # code

* = $1001
basic_start:
!wo basic_stop
!wo 0
!pet $9e, "4109", 0
basic_stop:
!wo 0

restart:
    lda #TOK_CLEAR
    jsr k_chrout

    ; reset score
    lda #0
    sta score
    sta scoreh

    jsr clear_area

    ; add first numbers
    jsr add_tile
    jsr add_tile

redraw:
    lda #TOK_HOME
    jsr k_chrout

    lda scoreh
    ldx score
    jsr b_linprt

    ldy #0
draw_loop_y:
    lda #TOK_CR
    jsr k_chrout
draw_loop_x:
    ldx area,y
    lda gfx,x
    jsr k_chrout
    iny
    cpy #AREA_SIZE
    beq ++
    tya
    and #3
    bne draw_loop_x
    beq draw_loop_y
++
get_input:
    jsr k_getinkbd
    ldx #LAST_KEY
-   cmp input_tbl_in,x
    beq get_input_found
    dex
    bpl -
    bmi get_input   ; bra
get_input_found:
    ; x = LAST_KEY -> restart
    cpx #LAST_KEY
    bne +
    jmp restart
+
    ; x = 0..3 -> move
    stx cmd

process_area:
    ; orientate the area so that each move is as if it was to the left
    jsr cond_rotate
    jsr cond_mirror

    ; process each row
    ldx #0
    stx areachanges

    ; remove zero elements
process_row:
    ldy #AREA_W + 1 ; add 0 terminator
    lda #0
-   sta newrow - 1,y
    dey
    bne -

    stx tmp

-   lda area,x
    beq +
    sta newrow,y
    iny
+
    inx
    txa
    and #3
    bne -

    ; merge same elements
    ldy #0
    ldx #0
-   lda newrow,y
    beq +++
    cmp newrow + 1,y
    bne ++
    iny
    ;sec
    adc #0
    pha
    adc score
    sta score
    bcc +
    inc scoreh
+   pla
++  sta newrow2,x
    iny
    inx
    bpl -   ; bra
+++
    ;lda #0
-   cpx #AREA_W
    beq +
    sta newrow2,x
    inx
    bne -   ; bra
+
    ; copy row back to area
    ; check if moved
    ldx tmp
    tay ; 0
-   lda newrow2,y
    cmp area,x
    beq +
    sta area,x
    inc areachanges
+   inx
    iny
    cpy #AREA_W
    bne -

    cpx #AREA_SIZE
    bne process_row

    ; rows done, restore area orientation
    jsr cond_mirror
    jsr cond_rotate

    ; check if moved
    lda areachanges
    bne +
    ; no changes, done
    jmp redraw
+

    ; check win
    ldx #AREA_SIZE - 1
    lda #$b
--  cmp area,x
    bne +
    ; won, clear area
    jsr clear_area
    ; ... and set B
    lda #$b
    sta area,x
    jmp redraw
+   dex
    bpl --

    jsr add_tile
    jmp redraw


; # subroutines

clear_area:
    ldy #AREA_SIZE - 1
    lda #0
-   sta area,y
    dey
    bpl -
    rts

cond_rotate:
    lda cmd
    cmp #2
    bcc +
    ldy #AREA_SIZE - 1
-   ldx rotate_tbl,y
    jsr swap_x_y
    bpl -
+   rts

cond_mirror:
    lda cmd
    lsr
    bcc +
    ldy #AREA_SIZE - 1
-   ldx mirror_tbl,y
    jsr swap_x_y
    bpl -
+   rts

swap_x_y:
    lda area,x
    sta tmp
    lda area,y
    sta area,x
    lda tmp
    sta area,y
    dey
    rts

add_tile:
    ; count empty
    ldx #AREA_SIZE - 1
    lda #0
    tay
-   cmp area,x
    bne +
    iny
+   dex
    bpl -

    cpy #0
    beq +++  ; game over

    ; get random position
    tya
    jsr get_random_a
    tay

    ; find position
    ldx #AREA_SIZE - 1
-   lda area,x
    bne +
    dey
    bmi ++
+   dex
    bpl -
++
    ; get random value
    ldy #2
    lda #100
    jsr get_random_a
    cmp #90
    bcs +
    dey
+   tya
    sta area,x

+++
    rts


; a <- a * rnd(1)
get_random_a:
    stx get_random_smc_x
    sty get_random_smc_y
    pha
    jsr k_rnd_1_to_fac
    jsr b_rftoa
    pla
    tay
    jsr b_y_to_fac
    jsr b_mul_ay_fac
    jsr b_intidx
    tya
get_random_smc_x = * + 1
    ldx #0
get_random_smc_y = * + 1
    ldy #0
    rts


; # data

rotate_tbl:
!by $0, $4, $8, $c
!by $4, $5, $9, $d
!by $8, $9, $a, $e
!by $c, $d, $e, $f

mirror_tbl:
!by $3, $2, $2, $3
!by $7, $6, $6, $7
!by $b, $a, $a, $b
!by $f, $e, $e, $f

input_tbl_in:
; NOTE: these are in a specific order!
; The code uses the following index bits:
;   bit 0: mirror
;   bit 1: rotate
; ... and assumes restart is last.
!raw "ADWSR"

gfx:
!by $71
!scr "123456789AB"
